// Copyright (c) 2019, QuantStack and Mamba Contributors
//
// Distributed under the terms of the BSD 3-Clause License.
//
// The full license is in the file LICENSE, distributed with this software.

#ifndef MAMBA_CORE_COMMON_TYPES_HPP
#define MAMBA_CORE_COMMON_TYPES_HPP

namespace mamba
{
    enum class log_level
    {
        trace,
        debug,
        info,
        warn,
        err,
        critical,
        off
    };
}

#endif
