// Copyright (c) 2019, QuantStack and Mamba Contributors
//
// Distributed under the terms of the BSD 3-Clause License.
//
// The full license is in the file LICENSE, distributed with this software.

#ifndef MAMBA_API_CONFIG_HPP
#define MAMBA_API_CONFIG_HPP


namespace mamba
{
    void config_describe();

    void config_list();

    void config_sources();
}

#endif
